// Copyright 2023-2024 The Khronos Group Inc.
// Copyright 2023-2024 Valve Corporation
// Copyright 2023-2024 LunarG, Inc.
//
// SPDX-License-Identifier: Apache-2.0

// This file exists to ensure the GN build checks our header files

#include <vulkan/layer/vk_layer_settings.h>
#include <vulkan/layer/vk_layer_settings.hpp>
#include <vulkan/utility/vk_dispatch_table.h>
#include <vulkan/utility/vk_concurrent_unordered_map.hpp>
#include <vulkan/utility/vk_format_utils.h>
#include <vulkan/utility/vk_safe_struct.hpp>
#include <vulkan/utility/vk_safe_struct_utils.hpp>
#include <vulkan/utility/vk_small_containers.hpp>
#include <vulkan/utility/vk_sparse_range_map.hpp>
#include <vulkan/utility/vk_struct_helper.hpp>
#include <vulkan/vk_enum_string_helper.h>
